@@ -22,7 +22,9 @@ public:
                                               const char& delim,
                                               int maxParts = 0);
   static time_t StringToTime(std::string timeString);
+  static std::string TimeToString(time_t time);
   static std::string ltrim(std::string str, const std::string chars = "\t\n\v\f\r _");
+  static std::string rtrim(std::string str, const std::string chars = "\t\n\v\f\r ");
   static int GetIDDirty(std::string str);
   static int GetChannelId(const char* strChannelName);
   static int stoiDefault(std::string str, int i);
@@ -30,4 +32,6 @@ public:
   static std::string ReplaceAll(std::string str,
                                 const std::string& search,
                                 const std::string& replace);
+  static std::string Replace(std::string str, const std::string& from, const std::string& to);
+  static std::string GenerateUuid();
 };
