@@ -1,5 +1,5 @@
 /*
- *      Copyright (C) 2019 flubshi
+ *      Copyright (C) 2019 - 2021 flubshi
  *      https://github.com/flubshi
  *
  *  This Program is free software; you can redistribute it and/or modify
@@ -59,8 +59,11 @@ string WaipuData::HttpRequest(const string& action, const string& url, const str
     curl.AddHeader(header.first, header.second);
   }
 
-  //curl.AddHeader("User-Agent", WAIPU_USER_AGENT);
-  curl.AddHeader("Authorization", "Bearer " + m_apiToken.accessToken);
+  curl.AddHeader("Authorization", "Bearer " + m_accessToken.getToken());
+
+
+  curl.AddHeader("User-Agent", WAIPU_USER_AGENT);
+  kodi::Log(ADDON_LOG_DEBUG, "HTTP User-Agent: %s.", WAIPU_USER_AGENT.c_str());
 
   return HttpRequestToCurl(curl, action, url, postData, statusCode);
 }
@@ -134,56 +137,46 @@ bool WaipuData::ApiLogin()
 
 bool WaipuData::ParseAccessToken(void)
 {
-  std::vector<std::string> jwt_arr = Utils::SplitString(m_apiToken.accessToken, '.', 3);
-  if (jwt_arr.size() == 3)
+  if(!m_accessToken.isInitialized() || m_accessToken.isExpired())
   {
-    kodi::Log(ADDON_LOG_DEBUG, "[jwt] middle: %s", jwt_arr.at(1).c_str());
-    string jwt_payload = base64_decode(jwt_arr.at(1));
-    kodi::Log(ADDON_LOG_DEBUG, "[jwt] payload: %s", jwt_payload.c_str());
-
-    Document jwt_doc;
-    jwt_doc.Parse(jwt_payload.c_str());
-
-    if (jwt_doc.HasParseError())
-    {
-      m_login_status = WAIPU_LOGIN_STATUS::UNKNOWN;
-      kodi::Log(ADDON_LOG_ERROR, "[jwt_doc] ERROR: error while parsing json");
-      return false;
-    }
-
-    m_userhandle = jwt_doc["userHandle"].GetString();
-    kodi::Log(ADDON_LOG_DEBUG, "[jwt] userHandle: %s", m_userhandle.c_str());
-    // generate the license
-    string license_plain = "{\"merchant\" : \"exaring\", \"sessionId\" : \"default\", "
-                           "\"userId\" : \"" +
-                           m_userhandle + "\"}";
-    kodi::Log(ADDON_LOG_DEBUG, "[jwt] license_plain: %s", license_plain.c_str());
-    m_license = base64_encode(license_plain.c_str(), license_plain.length());
-    kodi::Log(ADDON_LOG_DEBUG, "[jwt] license: %s", m_license.c_str());
-    // get user channels
-    m_user_channels_sd.clear();
-    m_user_channels_hd.clear();
-    for (const auto& user_channel : jwt_doc["userAssets"]["channels"]["SD"].GetArray())
-    {
-      string user_channel_s = user_channel.GetString();
-      kodi::Log(ADDON_LOG_DEBUG, "[jwt] SD channel: %s", user_channel_s.c_str());
-      m_user_channels_sd.push_back(user_channel_s);
-    }
-    for (const auto& user_channel : jwt_doc["userAssets"]["channels"]["HD"].GetArray())
-    {
-      string user_channel_s = user_channel.GetString();
-      m_user_channels_hd.push_back(user_channel_s);
-      kodi::Log(ADDON_LOG_DEBUG, "[jwt] HD channel: %s", user_channel_s.c_str());
-    }
-    if(jwt_doc["userAssets"].HasMember("instantRestart")){
-	m_account_replay_allowed = jwt_doc["userAssets"]["instantRestart"].GetBool();
-	kodi::Log(ADDON_LOG_DEBUG, "[jwt] Account InstantStart: %i", m_account_replay_allowed);
-    }
-    if(jwt_doc["userAssets"].HasMember("hoursRecording")){
-	m_account_hours_recording = jwt_doc["userAssets"]["hoursRecording"].GetInt();
-	kodi::Log(ADDON_LOG_DEBUG, "[jwt] Account HoursReording: %i", m_account_hours_recording);
-    }
+    m_login_status = WAIPU_LOGIN_STATUS::UNKNOWN;
+    kodi::Log(ADDON_LOG_ERROR, "[jwt_doc] ERROR: error while parsing json (error/expired)");
+    return false;
   }
+
+  m_userhandle = m_accessToken.parsedToken["userHandle"].GetString();
+  kodi::Log(ADDON_LOG_DEBUG, "[jwt] userHandle: %s", m_userhandle.c_str());
+  // generate the license
+  string license_plain = "{\"merchant\" : \"exaring\", \"sessionId\" : \"default\", "
+                         "\"userId\" : \"" +
+                         m_userhandle + "\"}";
+  kodi::Log(ADDON_LOG_DEBUG, "[jwt] license_plain: %s", license_plain.c_str());
+  m_license = base64_encode(license_plain.c_str(), license_plain.length());
+  kodi::Log(ADDON_LOG_DEBUG, "[jwt] license: %s", m_license.c_str());
+  // get user channels
+  m_user_channels_sd.clear();
+  m_user_channels_hd.clear();
+  for (const auto& user_channel : m_accessToken.parsedToken["userAssets"]["channels"]["SD"].GetArray())
+  {
+    string user_channel_s = user_channel.GetString();
+    kodi::Log(ADDON_LOG_DEBUG, "[jwt] SD channel: %s", user_channel_s.c_str());
+    m_user_channels_sd.push_back(user_channel_s);
+  }
+  for (const auto& user_channel : m_accessToken.parsedToken["userAssets"]["channels"]["HD"].GetArray())
+  {
+    string user_channel_s = user_channel.GetString();
+    m_user_channels_hd.push_back(user_channel_s);
+    kodi::Log(ADDON_LOG_DEBUG, "[jwt] HD channel: %s", user_channel_s.c_str());
+  }
+  if(m_accessToken.parsedToken["userAssets"].HasMember("instantRestart")){
+    m_account_replay_allowed = m_accessToken.parsedToken["userAssets"]["instantRestart"].GetBool();
+    kodi::Log(ADDON_LOG_DEBUG, "[jwt] Account InstantStart: %i", m_account_replay_allowed);
+  }
+  if(m_accessToken.parsedToken["userAssets"].HasMember("hoursRecording")){
+    m_account_hours_recording = m_accessToken.parsedToken["userAssets"]["hoursRecording"].GetInt();
+    kodi::Log(ADDON_LOG_DEBUG, "[jwt] Account HoursReording: %i", m_account_hours_recording);
+  }
+
   m_login_status = WAIPU_LOGIN_STATUS::OK;
   return true;
 }
@@ -196,8 +189,8 @@ bool WaipuData::WaipuLogin()
   time_t currTime;
   time(&currTime);
   kodi::Log(ADDON_LOG_DEBUG, "[token] current time %i", currTime);
-  kodi::Log(ADDON_LOG_DEBUG, "[token] expire  time %i", m_apiToken.expires);
-  if (!m_apiToken.accessToken.empty() && (m_apiToken.expires - 20 * 60) > currTime)
+  kodi::Log(ADDON_LOG_DEBUG, "[token] expire  time %i", m_accessToken.getExp());
+  if (m_accessToken.isInitialized() && !m_accessToken.isExpired(20 * 60))
   {
     // API token exists and is valid, more than x in future
     kodi::Log(ADDON_LOG_DEBUG, "[login check] old token still valid");
@@ -205,27 +198,28 @@ bool WaipuData::WaipuLogin()
   }
 
   ostringstream dataStream;
-  if (!m_apiToken.refreshToken.empty())
+  if (m_refreshToken.isInitialized() && !m_refreshToken.isExpired())
   {
     // Since the refresh token is valid for a long time, we do not check expiration for now
     // refresh API token
-    dataStream << "refresh_token=" << Utils::UrlEncode(m_apiToken.refreshToken)
-               << "&grant_type=refresh_token";
+    dataStream << "refresh_token=" << Utils::UrlEncode(m_refreshToken.getToken())
+               << "&grant_type=refresh_token"
+               << "&waipu_device_id=" << m_device_id;
     kodi::Log(ADDON_LOG_DEBUG, "[login check] Login-Request (refresh): %s;", dataStream.str().c_str());
   }
   else
   {
     // get API by login user/pw
     dataStream << "username=" << Utils::UrlEncode(m_username)
-               << "&password=" << Utils::UrlEncode(m_password) << "&grant_type=password";
-    kodi::Log(ADDON_LOG_DEBUG, "[login check] Login-Request (user/pw): %s;",
-              dataStream.str().c_str());
+               << "&password=" << Utils::UrlEncode(m_password)
+               << "&grant_type=password"
+               << "&waipu_device_id=" << m_device_id;
+    kodi::Log(ADDON_LOG_DEBUG, "[login check] Login-Request (user/pw)");
   }
   string jsonString;
   // curl request
   Curl curl;
   int statusCode = 0;
-  //curl.AddHeader("User-Agent", WAIPU_USER_AGENT);
   curl.AddHeader("Authorization", "Basic YW5kcm9pZENsaWVudDpzdXBlclNlY3JldA==");
   curl.AddHeader("Content-Type", "application/x-www-form-urlencoded");
   jsonString = HttpRequestToCurl(curl, "POST", "https://auth.waipu.tv/oauth/token",
@@ -243,6 +237,11 @@ bool WaipuData::WaipuLogin()
   }
   else if (statusCode == 401)
   {
+    if (m_refreshToken.isInitialized() && !m_refreshToken.isExpired())
+    {
+      // we used invalid refresh token, delete it
+      m_refreshToken = JWT();
+    }
     // invalid credentials
     m_login_status = WAIPU_LOGIN_STATUS::INVALID_CREDENTIALS;
     return false;
@@ -274,12 +273,15 @@ bool WaipuData::WaipuLogin()
       return false;
     }
 
-    m_apiToken.accessToken = doc["access_token"].GetString();
-    kodi::Log(ADDON_LOG_DEBUG, "[login check] accessToken: %s;", m_apiToken.accessToken.c_str());
-    m_apiToken.refreshToken = doc["refresh_token"].GetString();
-    kodi::Log(ADDON_LOG_DEBUG, "[login check] refreshToken: %s;", m_apiToken.refreshToken.c_str());
-    m_apiToken.expires = currTime + doc["expires_in"].GetUint64();
-    kodi::Log(ADDON_LOG_DEBUG, "[login check] expires: %i;", m_apiToken.expires);
+    m_accessToken = JWT(doc["access_token"].GetString());
+    kodi::Log(ADDON_LOG_DEBUG, "[login check] accessToken: %s;", m_accessToken.getToken().c_str());
+    string refresh_token = doc["refresh_token"].GetString();
+    if (!refresh_token.empty())
+    {
+      m_refreshToken = JWT(refresh_token);
+      kodi::SetSettingString("refresh_token", refresh_token);
+      kodi::Log(ADDON_LOG_DEBUG, "[login check] refreshToken: %s;", refresh_token.c_str());
+    }
 
     return ParseAccessToken();
   }
@@ -294,7 +296,7 @@ bool WaipuData::O2Login()
   time_t currTime;
   time(&currTime);
 
-  if (!m_apiToken.accessToken.empty() && (m_apiToken.expires - 10 * 60) > currTime)
+  if (m_accessToken.isInitialized() && !m_accessToken.isExpired(10 * 60))
   {
     // API token exists and is valid, more than x in future
     kodi::Log(ADDON_LOG_DEBUG, "[login check] old token still valid");
@@ -306,7 +308,6 @@ bool WaipuData::O2Login()
   // curl request
   Curl curl;
   int statusCode = 0;
-  //curl.AddHeader("User-Agent", WAIPU_USER_AGENT);
   curl.AddHeader("authority", "o2api.waipu.tv");
   string respForm =
       HttpRequestToCurl(curl, "GET",
@@ -376,12 +377,12 @@ bool WaipuData::O2Login()
     return false;
   }
 
-  m_apiToken.accessToken = cookie;
+  m_accessToken = JWT(cookie);
   kodi::Log(ADDON_LOG_DEBUG, "[login O2] access_token: %s;", cookie.c_str());
-  m_apiToken.refreshToken = "";
+  m_refreshToken = JWT();
   kodi::Log(ADDON_LOG_DEBUG, "[login check] refreshToken: empty");
-  m_apiToken.expires = currTime + 3600; // expires after 1h; TODO: find real value
-  kodi::Log(ADDON_LOG_DEBUG, "[login check] expires: %i;", m_apiToken.expires);
+  //m_apiToken.expires = currTime + 3600; // expires after 1h; TODO: find real value
+  //kodi::Log(ADDON_LOG_DEBUG, "[login check] expires: %i;", m_apiToken.expires);
 
   return ParseAccessToken();
 }
@@ -390,11 +391,8 @@ bool WaipuData::RefreshDeviceCapabiltiesToken()
 {
   kodi::Log(ADDON_LOG_DEBUG, "%s - Creating the waipu.tv PVR add-on", __FUNCTION__);
 
-  time_t currTime;
-  time(&currTime);
-  kodi::Log(ADDON_LOG_DEBUG, "[device token] current time %i", currTime);
-  kodi::Log(ADDON_LOG_DEBUG, "[device token] expire  time %i", m_deviceCapabilitiesToken.expires);
-  if (!m_deviceCapabilitiesToken.token.empty() && (m_deviceCapabilitiesToken.expires - 5 * 60) > currTime)
+  kodi::Log(ADDON_LOG_DEBUG, "[device token] expire time %i", m_deviceCapabilitiesToken.getExp());
+  if (m_deviceCapabilitiesToken.isInitialized() && !m_deviceCapabilitiesToken.isExpired(5*60))
   {
     // device token exists and is valid, more than x in future
     kodi::Log(ADDON_LOG_DEBUG, "[device token] old token still valid, no need to refresh");
@@ -438,15 +436,8 @@ bool WaipuData::RefreshDeviceCapabiltiesToken()
 
   if(deviceTokenDoc.HasMember("token"))
   {
-    m_deviceCapabilitiesToken.token = deviceTokenDoc["token"].GetString();
-    kodi::Log(ADDON_LOG_DEBUG, "[X-Device-Token] discovered token: %s", m_deviceCapabilitiesToken.token.c_str());
-
-    if(deviceTokenDoc.HasMember("expiresIn")){
-	m_deviceCapabilitiesToken.expires = currTime + deviceTokenDoc["expiresIn"].GetUint64();
-	kodi::Log(ADDON_LOG_DEBUG, "[X-Device-Token] expires: %i;", m_deviceCapabilitiesToken.expires);
-    }else{
-	m_apiToken.expires = currTime + 5 * 60;
-    }
+    m_deviceCapabilitiesToken = JWT(deviceTokenDoc["token"].GetString());
+    kodi::Log(ADDON_LOG_DEBUG, "[X-Device-Token] discovered token: %s", m_deviceCapabilitiesToken.getToken().c_str());
     return true;
   }
 
@@ -459,6 +450,10 @@ ADDON_STATUS WaipuData::Create()
 {
   kodi::Log(ADDON_LOG_DEBUG, "%s - Creating the waipu.tv PVR add-on", __FUNCTION__);
 
+  // set User-Agent
+  std::string ua = kodi::network::GetUserAgent();
+  WAIPU_USER_AGENT = Utils::Replace(ua, " ", std::string(" pvr.waipu/").append(STR(IPTV_VERSION)).append(" "));
+
   ReadSettings();
 
   ADDON_STATUS curStatus = ADDON_STATUS_UNKNOWN;
@@ -507,8 +502,18 @@ void WaipuData::ReadSettings(void)
 
   m_username = kodi::GetSettingString("username");
   m_password = kodi::GetSettingString("password");
-  m_protocol = kodi::GetSettingString("protocol", "dash");
+  m_protocol = kodi::GetSettingString("protocol", "auto");
   m_provider = kodi::GetSettingEnum<WAIPU_PROVIDER>("provider_select", WAIPU_PROVIDER_WAIPU);
+  m_refreshToken = JWT(kodi::GetSettingString("refresh_token", ""));
+
+  m_device_id = kodi::GetSettingString("device_id_uuid4");
+  if (m_device_id.empty())
+  {
+    m_device_id = Utils::GenerateUuid();
+    kodi::SetSettingString("device_id_uuid4", m_device_id);
+    // new device id -> force new login
+    m_refreshToken = JWT();
+  }
 
   kodi::Log(ADDON_LOG_DEBUG, "End Readsettings");
 }
@@ -523,10 +528,10 @@ ADDON_STATUS WaipuData::SetSetting(const std::string& settingName,
     {
       m_username = username;
       m_login_failed_counter = 0;
+      kodi::SetSettingString("refresh_token", "");
       return ADDON_STATUS_NEED_RESTART;
     }
   }
-
   else if (settingName == "password")
   {
     std::string password = settingValue.GetString();
@@ -534,6 +539,7 @@ ADDON_STATUS WaipuData::SetSetting(const std::string& settingName,
     {
       m_login_failed_counter = 0;
       m_password = password;
+      kodi::SetSettingString("refresh_token", "");
       return ADDON_STATUS_NEED_RESTART;
     }
   }
@@ -555,13 +561,14 @@ ADDON_STATUS WaipuData::SetSetting(const std::string& settingName,
     {
       m_login_failed_counter = 0;
       m_provider = tmpProvider;
+      kodi::SetSettingString("refresh_token", "");
       return ADDON_STATUS_NEED_RESTART;
     }
   }else if (settingName.rfind("streaming_capabilities_", 0) == 0)
   {
     // settings name begins with "streaming_capabilities_"
     // reset capabilities to force refresh
-    m_deviceCapabilitiesToken.token = "";
+    m_deviceCapabilitiesToken = JWT();
   }
 
   return ADDON_STATUS_OK;
@@ -605,7 +612,8 @@ std::string WaipuData::GetLicense(void)
 
 void WaipuData::SetStreamProperties(std::vector<kodi::addon::PVRStreamProperty>& properties,
                                     const std::string& url,
-                                    bool realtime, bool playTimeshiftBuffer)
+                                    bool realtime, bool playTimeshiftBuffer,
+                                    const std::string& protocol)
 {
   kodi::Log(ADDON_LOG_DEBUG, "[PLAY STREAM] url: %s", url.c_str());
 
@@ -613,7 +621,7 @@ void WaipuData::SetStreamProperties(std::vector<kodi::addon::PVRStreamProperty>&
   properties.emplace_back(PVR_STREAM_PROPERTY_INPUTSTREAM, "inputstream.adaptive");
   properties.emplace_back(PVR_STREAM_PROPERTY_ISREALTIMESTREAM, realtime ? "true" : "false");
 
-  if (m_protocol == "dash")
+  if (protocol == "dash")
   {
     // MPEG DASH
     kodi::Log(ADDON_LOG_DEBUG, "[PLAY STREAM] dash");
@@ -635,7 +643,7 @@ void WaipuData::SetStreamProperties(std::vector<kodi::addon::PVRStreamProperty>&
 
     properties.emplace_back("inputstream.adaptive.manifest_update_parameter", "full");
   }
-  else if (m_protocol == "hls")
+  else if (protocol == "hls")
   {
     // HLS
     kodi::Log(ADDON_LOG_DEBUG, "[PLAY STREAM] hls");
@@ -645,7 +653,7 @@ void WaipuData::SetStreamProperties(std::vector<kodi::addon::PVRStreamProperty>&
   }
   else
   {
-    kodi::Log(ADDON_LOG_ERROR, "[SetStreamProperties] called with invalid protocol '%s'", m_protocol.c_str());
+    kodi::Log(ADDON_LOG_ERROR, "[SetStreamProperties] called with invalid protocol '%s'", protocol.c_str());
   }
 }
 
@@ -848,13 +856,32 @@ PVR_ERROR WaipuData::GetChannels(bool radio, kodi::addon::PVRChannelsResultSet&
 PVR_ERROR WaipuData::GetChannelStreamProperties(
     const kodi::addon::PVRChannel& channel, std::vector<kodi::addon::PVRStreamProperty>& properties)
 {
-
-  string strUrl = GetChannelStreamUrl(channel.GetUniqueId(), m_protocol, "");
-  kodi::Log(ADDON_LOG_DEBUG, "Stream URL -> %s", strUrl.c_str());
   PVR_ERROR ret = PVR_ERROR_FAILED;
+  string protocol = m_protocol;
+  if (protocol == "auto")
+  {
+    // use hls where possible, fallback to dash
+    protocol = "dash";
+    for (const auto& thisChannel : m_channels)
+    {
+      if (thisChannel.iUniqueId == channel.GetUniqueId())
+      {
+        if (m_hls_allowlist.contains(thisChannel.waipuID))
+        {
+          protocol = "hls";
+          break;
+        }
+      }
+    }
+    kodi::Log(ADDON_LOG_DEBUG, "protocol auto select: %s", protocol.c_str());
+  }
+
+  string strUrl = GetChannelStreamUrl(channel.GetUniqueId(), protocol, "");
+  kodi::Log(ADDON_LOG_DEBUG, "Stream URL -> %s", strUrl.c_str());
+
   if (!strUrl.empty())
   {
-    SetStreamProperties(properties, strUrl, true, false);
+    SetStreamProperties(properties, strUrl, true, false, protocol);
     ret = PVR_ERROR_NO_ERROR;
   }
   return ret;
@@ -886,7 +913,7 @@ string WaipuData::GetChannelStreamUrl(int uniqueId, const string& protocol, cons
       postData += "}}";
       kodi::Log(ADDON_LOG_DEBUG, "[GetStreamURL] Post data: %s", postData.c_str());
 
-      string jsonStreamURL = HttpPost("https://stream-url-provider.waipu.tv/api/stream-url", postData, {{"Content-Type", "application/vnd.streamurlprovider.stream-url-request-v1+json"}, {"X-Device-Token", m_deviceCapabilitiesToken.token.c_str()}});
+      string jsonStreamURL = HttpPost("https://stream-url-provider.waipu.tv/api/stream-url", postData, {{"Content-Type", "application/vnd.streamurlprovider.stream-url-request-v1+json"}, {"X-Device-Token", m_deviceCapabilitiesToken.getToken().c_str()}});
 
       Document streamURLDoc;
       streamURLDoc.Parse(jsonStreamURL.c_str());
@@ -969,15 +996,8 @@ PVR_ERROR WaipuData::GetEPGForChannel(int channelUid,
     if (myChannel.iUniqueId != channelUid)
       continue;
 
-    char startTime[100];
-    std::tm* pstm = std::localtime(&start);
-    // 2019-01-20T23:59:59
-    std::strftime(startTime, 32, "%Y-%m-%dT%H:%M:%S", pstm);
-
-    char endTime[100];
-    std::tm* petm = std::localtime(&end);
-    // 2019-01-20T23:59:59
-    std::strftime(endTime, 32, "%Y-%m-%dT%H:%M:%S", petm);
+    string startTime = Utils::TimeToString(start);
+    string endTime = Utils::TimeToString(end);
 
     string jsonEpg =
         HttpGet("https://epg.waipu.tv/api/channels/" + myChannel.waipuID +
@@ -1004,7 +1024,6 @@ PVR_ERROR WaipuData::GetEPGForChannel(int channelUid,
     for (const auto& epgData : epgDoc["result"].GetArray())
     {
       kodi::addon::PVREPGTag tag;
-      WaipuEPGEntry epgEntry;
 
       // generate a unique boadcast id
       string epg_bid = epgData["id"].GetString();
@@ -1012,31 +1031,28 @@ PVR_ERROR WaipuData::GetEPGForChannel(int channelUid,
       int dirtyID = Utils::GetIDDirty(epg_bid);
       kodi::Log(ADDON_LOG_DEBUG, "[epg] epg_bid dirty: %i;", dirtyID);
       tag.SetUniqueBroadcastId(dirtyID);
-      epgEntry.iUniqueBroadcastId = dirtyID;
 
       // channel ID
       tag.SetUniqueChannelId(myChannel.iUniqueId);
-      epgEntry.iUniqueChannelId = myChannel.iUniqueId;
 
-      // add streamUrlProvider if it is video on demand
+      /*// add streamUrlProvider if it is video on demand
       if (myChannel.tvfuse && epgData.HasMember("streamUrlProvider") && !epgData["streamUrlProvider"].IsNull())
       {
         string streamUrlProvider = epgData["streamUrlProvider"].GetString();
         kodi::Log(ADDON_LOG_DEBUG, "[epg] streamUrlProvider: %s;", streamUrlProvider.c_str());
-        epgEntry.streamUrlProvider = streamUrlProvider;
-      }
+      }*/
+
+      unsigned int flags = EPG_TAG_FLAG_UNDEFINED;
 
       // is recordable
       bool isRecordable = !epgData["recordingForbidden"].GetBool();
       kodi::Log(ADDON_LOG_DEBUG, "[epg] recordable: %i;", isRecordable);
-      epgEntry.isRecordable = isRecordable;
+      if (isRecordable){flags |= EPG_TAG_FLAG_IS_RECORDABLE_WAIPU; }
 
       // instantRestartAllowed
       bool instantRestartAllowed = !epgData["instantRestartForbidden"].GetBool();
       kodi::Log(ADDON_LOG_DEBUG, "[epg] instantRestartAllowed: %i;", instantRestartAllowed);
-      epgEntry.instantRestartAllowed = instantRestartAllowed;
-
-      m_epgEntries.push_back(epgEntry);
+      if (isRecordable){flags |= EPG_TAG_FLAG_INSTANT_RESTART_ALLOWED_WAIPU; }
 
       // set title
       tag.SetTitle(epgData["title"].GetString());
@@ -1061,13 +1077,12 @@ PVR_ERROR WaipuData::GetEPGForChannel(int channelUid,
 
       // tag.SetIconPath(myTag.strIconPath);
 
-      tag.SetFlags(EPG_TAG_FLAG_UNDEFINED);
-
       // iSeriesNumber
       if (epgData.HasMember("season") && !epgData["season"].IsNull())
       {
         tag.SetSeriesNumber(
             Utils::stoiDefault(epgData["season"].GetString(), EPG_TAG_INVALID_SERIES_EPISODE));
+        flags |= EPG_TAG_FLAG_IS_SERIES;
       }
       else
       {
@@ -1083,7 +1098,6 @@ PVR_ERROR WaipuData::GetEPGForChannel(int channelUid,
       {
         tag.SetEpisodeNumber(EPG_TAG_INVALID_SERIES_EPISODE);
       }
-      tag.SetEpisodePartNumber(EPG_TAG_INVALID_SERIES_EPISODE);
 
       // episodeName
       if (epgData.HasMember("episodeTitle") && !epgData["episodeTitle"].IsNull())
@@ -1100,10 +1114,22 @@ PVR_ERROR WaipuData::GetEPGForChannel(int channelUid,
       // genre
       if (epgData.HasMember("genreDisplayName") && !epgData["genreDisplayName"].IsNull())
       {
-        tag.SetGenreType(EPG_GENRE_USE_STRING);
-        tag.SetGenreDescription(epgData["genreDisplayName"].GetString());
+        string genreStr = epgData["genreDisplayName"].GetString();
+        int genre = m_categories.Category(genreStr);
+        if (genre)
+        {
+          tag.SetGenreSubType(genre & 0x0F);
+          tag.SetGenreType(genre & 0xF0);
+        }
+        else
+        {
+          tag.SetGenreType(EPG_GENRE_USE_STRING);
+          tag.SetGenreSubType(0); /* not supported */
+          tag.SetGenreDescription(genreStr);
+        }
       }
 
+      tag.SetFlags(flags);
       results.Add(tag);
     }
   }
@@ -1128,17 +1154,7 @@ PVR_ERROR WaipuData::IsEPGTagRecordable(const kodi::addon::PVREPGTag& tag, bool&
     return PVR_ERROR_NO_ERROR;
   }
 
-  for (const auto& epgEntry : m_epgEntries)
-  {
-    if (epgEntry.iUniqueBroadcastId != tag.GetUniqueBroadcastId())
-      continue;
-    if (epgEntry.iUniqueChannelId != tag.GetUniqueChannelId())
-      continue;
-    isRecordable = epgEntry.isRecordable;
-    return PVR_ERROR_NO_ERROR;
-  }
-
-  isRecordable = false;
+  isRecordable = (tag.GetFlags() & EPG_TAG_FLAG_IS_RECORDABLE_WAIPU);
   return PVR_ERROR_NO_ERROR;
 }
 
@@ -1161,16 +1177,7 @@ PVR_ERROR WaipuData::IsEPGTagPlayable(const kodi::addon::PVREPGTag& tag, bool& i
   auto current_time = time(NULL);
   if (m_account_replay_allowed && current_time > tag.GetStartTime() && current_time < tag.GetEndTime())
   {
-      // tag is now running, check if epg tag allows replay
-      for (const auto& epgEntry : m_epgEntries)
-      {
-        if (epgEntry.iUniqueBroadcastId != tag.GetUniqueBroadcastId())
-          continue;
-        if (epgEntry.iUniqueChannelId != tag.GetUniqueChannelId())
-          continue;
-        isPlayable = epgEntry.instantRestartAllowed;
-        return PVR_ERROR_NO_ERROR;
-      }
+      isPlayable = (tag.GetFlags() & EPG_TAG_FLAG_INSTANT_RESTART_ALLOWED_WAIPU);
   }
 
   return PVR_ERROR_NO_ERROR;
@@ -1181,13 +1188,16 @@ PVR_ERROR WaipuData::GetEPGTagStreamProperties(
 {
   kodi::Log(ADDON_LOG_DEBUG, "[EPG TAG] play it...");
 
-  string strUrl = GetEPGTagURL(tag, m_protocol);
+  string protocol = m_protocol;
+  if (protocol == "auto") protocol = "dash";  //fallback to dash
+
+  string strUrl = GetEPGTagURL(tag, protocol);
   if (strUrl.empty())
   {
     return PVR_ERROR_FAILED;
   }
 
-  SetStreamProperties(properties, strUrl, true, true);
+  SetStreamProperties(properties, strUrl, true, true, protocol);
 
   return PVR_ERROR_NO_ERROR;
 }
@@ -1196,43 +1206,63 @@ string WaipuData::GetEPGTagURL(const kodi::addon::PVREPGTag& tag, const string&
 {
   ApiLogin();
 
-  for (const auto& epgEntry : m_epgEntries)
+  for (const auto& thisChannel : m_channels)
   {
-    if (epgEntry.iUniqueChannelId != tag.GetUniqueChannelId())
-      continue;
-    if (epgEntry.iUniqueBroadcastId != tag.GetUniqueBroadcastId())
-      continue;
-
-    string url = epgEntry.streamUrlProvider;
-    if (!url.empty())
+    if (thisChannel.iUniqueId == tag.GetUniqueChannelId())
     {
-      kodi::Log(ADDON_LOG_DEBUG, "play epgTAG -> %s", tag.GetTitle().c_str());
-      kodi::Log(ADDON_LOG_DEBUG, "play url -> %s", url.c_str());
+      string startTime = Utils::TimeToString(tag.GetStartTime());
+      string endTime = Utils::TimeToString(tag.GetEndTime());
 
-      string tag_resp = HttpGet(url);
-      kodi::Log(ADDON_LOG_DEBUG, "tag resp -> %s", tag_resp.c_str());
-
-      Document tagDoc;
-      tagDoc.Parse(tag_resp.c_str());
-      if (tagDoc.GetParseError())
+      string jsonEpg =
+	  HttpGet("https://epg.waipu.tv/api/channels/" + thisChannel.waipuID +
+	          "/programs?includeRunningAtStartTime=false&startTime=" + string(startTime) + "&stopTime=" + string(endTime));
+      kodi::Log(ADDON_LOG_DEBUG, "[epg-single-tag] %s", jsonEpg.c_str());
+      if (jsonEpg.size() == 0)
       {
-        kodi::Log(ADDON_LOG_ERROR, "[getEPGTagURL] ERROR: error while parsing json");
+        kodi::Log(ADDON_LOG_ERROR, "[epg-single-tag] empty server response");
         return "";
       }
-      kodi::Log(ADDON_LOG_DEBUG, "[tag] streams");
-      // check if streams there
-      if (tagDoc.HasMember("player") && tagDoc["player"].HasMember("mpd"))
+      jsonEpg = "{\"result\": " + jsonEpg + "}";
+
+      Document epgDoc;
+      epgDoc.Parse(jsonEpg.c_str());
+
+      if (epgDoc.GetParseError() || epgDoc["result"].Size() == 0 || !epgDoc["result"][0].HasMember("streamUrlProvider") || epgDoc["result"][0]["streamUrlProvider"].IsNull())
       {
-        string mpdUrl = tagDoc["player"]["mpd"].GetString();
-        kodi::Log(ADDON_LOG_DEBUG, "mpd url -> %s", mpdUrl.c_str());
-        return mpdUrl;
+        // fallback to replay playback
+        kodi::Log(ADDON_LOG_DEBUG, "[play epg tag] streamUrlProvider not found -> fallback to replay!");
+        string startTime = std::to_string(tag.GetStartTime());
+        return GetChannelStreamUrl(tag.GetUniqueChannelId(), protocol, startTime);
+      }
+
+      string url = epgDoc["result"][0]["streamUrlProvider"].GetString();
+
+      if (!url.empty())
+      {
+        kodi::Log(ADDON_LOG_DEBUG, "play url -> %s", url.c_str());
+
+        string tag_resp = HttpGet(url);
+        kodi::Log(ADDON_LOG_DEBUG, "tag resp -> %s", tag_resp.c_str());
+
+        Document tagDoc;
+        tagDoc.Parse(tag_resp.c_str());
+        if (tagDoc.GetParseError())
+        {
+          kodi::Log(ADDON_LOG_ERROR, "[getEPGTagURL] ERROR: error while parsing json");
+          return "";
+        }
+        kodi::Log(ADDON_LOG_DEBUG, "[tag] streams");
+        // check if streams there
+        if (tagDoc.HasMember("player") && tagDoc["player"].HasMember("mpd"))
+        {
+          string mpdUrl = tagDoc["player"]["mpd"].GetString();
+          kodi::Log(ADDON_LOG_DEBUG, "mpd url -> %s", mpdUrl.c_str());
+          return mpdUrl;
+        }
       }
     }
-
-    // fallback to replay playback
-    string startTime = std::to_string(tag.GetStartTime());
-    return GetChannelStreamUrl(tag.GetUniqueChannelId(), protocol, startTime);
   }
+  kodi::Log(ADDON_LOG_DEBUG, "[play epg tag] channel or tag not found!");
   return "";
 }
 
@@ -1372,9 +1402,19 @@ PVR_ERROR WaipuData::GetRecordings(bool deleted, kodi::addon::PVRRecordingsResul
     // genre
     if (epgData.HasMember("genreDisplayName") && !epgData["genreDisplayName"].IsNull())
     {
-      tag.SetGenreType(EPG_GENRE_USE_STRING);
-      string genre = epgData["genreDisplayName"].GetString();
-      tag.SetGenreDescription(genre);
+      string genreStr = epgData["genreDisplayName"].GetString();
+      int genre = m_categories.Category(genreStr);
+      if (genre)
+      {
+        tag.SetGenreSubType(genre & 0x0F);
+        tag.SetGenreType(genre & 0xF0);
+      }
+      else
+      {
+        tag.SetGenreType(EPG_GENRE_USE_STRING);
+        tag.SetGenreSubType(0); /* not supported */
+        tag.SetGenreDescription(genreStr);
+      }
     }
 
     // epg mapping
@@ -1460,13 +1500,16 @@ PVR_ERROR WaipuData::GetRecordingStreamProperties(
 {
   kodi::Log(ADDON_LOG_DEBUG, "[recordings] play it...");
 
-  string strUrl = GetRecordingURL(recording, m_protocol);
+  string protocol = m_protocol;
+  if (protocol == "auto") protocol = "dash"; //fallback to dash
+
+  string strUrl = GetRecordingURL(recording, protocol);
   if (strUrl.empty())
   {
     return PVR_ERROR_FAILED;
   }
 
-  SetStreamProperties(properties, strUrl, true, false);
+  SetStreamProperties(properties, strUrl, true, false, protocol);
 
   return PVR_ERROR_NO_ERROR;
 }
@@ -1481,11 +1524,9 @@ void WaipuData::AddTimerType(std::vector<kodi::addon::PVRTimerType>& types, int
 
 PVR_ERROR WaipuData::GetTimerTypes(std::vector<kodi::addon::PVRTimerType>& types)
 {
-  //AddTimerType(types, 1, PVR_TIMER_TYPE_ATTRIBUTE_NONE);
-  AddTimerType(types, 1,
-               PVR_TIMER_TYPE_SUPPORTS_READONLY_DELETE | PVR_TIMER_TYPE_SUPPORTS_CHANNELS |
-                   PVR_TIMER_TYPE_SUPPORTS_START_TIME | PVR_TIMER_TYPE_SUPPORTS_END_TIME);
-  //AddTimerType(types, 2, PVR_TIMER_TYPE_IS_MANUAL);
+  AddTimerType(types, 1, PVR_TIMER_TYPE_REQUIRES_EPG_TAG_ON_CREATE);
+  AddTimerType(types, 2, PVR_TIMER_TYPE_REQUIRES_EPG_SERIES_ON_CREATE | PVR_TIMER_TYPE_IS_REPEATING );
+
   return PVR_ERROR_NO_ERROR;
 }
 
@@ -1520,6 +1561,8 @@ PVR_ERROR WaipuData::GetTimers(kodi::addon::PVRTimersResultSet& results)
   int recordings_count = 0;
   int timers_count = 0;
 
+  std::list<int> timerGroups;
+
   for (const auto& timer : timersDoc["result"].GetArray())
   {
     // skip not FINISHED entries
@@ -1543,13 +1586,15 @@ PVR_ERROR WaipuData::GetTimers(kodi::addon::PVRTimersResultSet& results)
       tag.SetState(PVR_TIMER_STATE_RECORDING);
     }
     tag.SetLifetime(0);
-    tag.SetTimerType(1); // not the best way to do it...
 
-    // set recording id
-    string rec_id = timer["id"].GetString();
-    tag.SetClientIndex(Utils::stoiDefault(rec_id, 0));
-    tag.SetEPGUid(Utils::stoiDefault(rec_id, 0));
+    const Value& epgData = timer["epgData"];
 
+    // set recording title
+    string rec_title = epgData["title"].GetString();
+    kodi::Log(ADDON_LOG_DEBUG, "[timers] Add: %s;", rec_title.c_str());
+    tag.SetTitle(rec_title);
+
+    int tag_channel;
     // channelid
     if (timer.HasMember("channelId") && !timer["channelId"].IsNull())
     {
@@ -1559,17 +1604,39 @@ PVR_ERROR WaipuData::GetTimers(kodi::addon::PVRTimersResultSet& results)
         WaipuChannel& myChannel = m_channels.at(iChannelPtr);
         if (myChannel.waipuID != channel_name)
           continue;
-        tag.SetClientChannelUid(myChannel.iUniqueId);
+        tag_channel = myChannel.iUniqueId;
+        tag.SetClientChannelUid(tag_channel);
         break;
       }
     }
 
-    const Value& epgData = timer["epgData"];
+    if (timer.HasMember("recordingGroup"))
+    {
 
-    // set recording title
-    string rec_title = epgData["title"].GetString();
-    kodi::Log(ADDON_LOG_DEBUG, "[timers] Add: %s;", rec_title.c_str());
-    tag.SetTitle(rec_title);
+	int group = timer["recordingGroup"].GetInt();
+	tag.SetRecordingGroup(group);
+	if (std::find(timerGroups.begin(), timerGroups.end(), group) == timerGroups.end())
+	{
+	  // add group
+	  kodi::addon::PVRTimer tagGroup;
+	  tagGroup.SetTimerType(2);
+	  tagGroup.SetTitle(rec_title);
+	  tagGroup.SetClientIndex(group);
+	  tagGroup.SetClientChannelUid(tag_channel);
+	  tag.SetRecordingGroup(group);
+	  kodi::Log(ADDON_LOG_DEBUG, "[add timer group] group: %i;", group);
+
+	  results.Add(tagGroup);
+	  timerGroups.push_back(group);
+	}
+    }
+
+    tag.SetTimerType(1);
+
+    // set recording id
+    string rec_id = timer["id"].GetString();
+    tag.SetClientIndex(Utils::stoiDefault(rec_id, 0));
+    tag.SetEPGUid(Utils::stoiDefault(rec_id, 0));
 
     // get recording time
     if (timer.HasMember("startTime") && !timer["startTime"].IsNull())
@@ -1617,13 +1684,30 @@ PVR_ERROR WaipuData::DeleteTimer(const kodi::addon::PVRTimer& timer, bool forceD
 {
   if (ApiLogin())
   {
-    int timer_id = timer.GetClientIndex();
-    string request_data = "{\"ids\":[\"" + to_string(timer_id) + "\"]}";
-    kodi::Log(ADDON_LOG_DEBUG, "[delete timer] req: %s;", request_data.c_str());
-    string deleted = HttpDelete("https://recording.waipu.tv/api/recordings", request_data.c_str(),{{"Content-Type", "application/vnd.waipu.pvr-recording-ids-v2+json"}});
-    kodi::Log(ADDON_LOG_DEBUG, "[delete timer] response: %s;", deleted.c_str());
-    kodi::addon::CInstancePVRClient::TriggerTimerUpdate();
-    return PVR_ERROR_NO_ERROR;
+    if (timer.GetTimerType() == 1)
+    {
+      // single tag
+      int timer_id = timer.GetClientIndex();
+      string request_data = "{\"ids\":[\"" + to_string(timer_id) + "\"]}";
+      kodi::Log(ADDON_LOG_DEBUG, "[delete single timer] req: %s;", request_data.c_str());
+      string deleted = HttpDelete("https://recording.waipu.tv/api/recordings", request_data.c_str(),{{"Content-Type", "application/vnd.waipu.pvr-recording-ids-v2+json"}});
+      kodi::Log(ADDON_LOG_DEBUG, "[delete single timer] response: %s;", deleted.c_str());
+      kodi::QueueNotification(QUEUE_INFO, "Recording", "Recording Deleted");
+      kodi::addon::CInstancePVRClient::TriggerRecordingUpdate();
+      kodi::addon::CInstancePVRClient::TriggerTimerUpdate();
+      return PVR_ERROR_NO_ERROR;
+    }else{
+      // delete record series
+      int groupID = timer.GetClientIndex();
+      string request_data = "{\"serialRecordings\":[{\"id\":" + to_string(groupID) + ",\"deleteFutureRecordings\":true,\"deleteFinishedRecordings\":false,\"deleteRunningRecordingss\":false}]}";
+      kodi::Log(ADDON_LOG_DEBUG, "[delete multi timer] req (group: %i): %s;", groupID, request_data.c_str());
+      string deleted = HttpPost("https://recording-scheduler.waipu.tv/api/delete-requests", request_data.c_str(),{{"Content-Type", "application/vnd.waipu.recording-scheduler-delete-serial-recordings-v1+json"}});
+      kodi::Log(ADDON_LOG_DEBUG, "[delete multi timer] response: %s;", deleted.c_str());
+      kodi::QueueNotification(QUEUE_INFO, "Recording", "Rule Deleted");
+      kodi::addon::CInstancePVRClient::TriggerRecordingUpdate();
+      kodi::addon::CInstancePVRClient::TriggerTimerUpdate();
+      return PVR_ERROR_NO_ERROR;
+    }
   }
   return PVR_ERROR_FAILED;
 }
@@ -1638,17 +1722,36 @@ PVR_ERROR WaipuData::AddTimer(const kodi::addon::PVRTimer& timer)
 
   if (ApiLogin())
   {
-    // {"programId":"_1051966761","channelId":"PRO7","startTime":"2019-02-03T18:05:00.000Z","stopTime":"2019-02-03T19:15:00.000Z"}
     for (const auto& channel : m_channels)
     {
       if (channel.iUniqueId != timer.GetClientChannelUid())
         continue;
-      string postData = "{\"programId\":\"_" + to_string(timer.GetEPGUid()) +
-                        "\",\"channelId\":\"" + channel.waipuID + "\"" + "}";
-      string recordResp = HttpPost("https://recording.waipu.tv/api/recordings", postData, {{"Content-Type", "application/vnd.waipu.start-recording-v2+json"}});
-      kodi::Log(ADDON_LOG_DEBUG, "[add timer] response: %s;", recordResp.c_str());
-      kodi::addon::CInstancePVRClient::TriggerTimerUpdate();
-      return PVR_ERROR_NO_ERROR;
+
+      if (timer.GetTimerType() == 1)
+      {
+        // record single element
+        kodi::Log(ADDON_LOG_DEBUG, "[add timer] Record single tag;");
+        // {"programId":"_1051966761","channelId":"PRO7","startTime":"2019-02-03T18:05:00.000Z","stopTime":"2019-02-03T19:15:00.000Z"}
+        string postData = "{\"programId\":\"_" + to_string(timer.GetEPGUid()) +
+                          "\",\"channelId\":\"" + channel.waipuID + "\"" + "}";
+        string recordResp = HttpPost("https://recording.waipu.tv/api/recordings", postData, {{"Content-Type", "application/vnd.waipu.start-recording-v2+json"}});
+        kodi::Log(ADDON_LOG_DEBUG, "[add timer] single response: %s;", recordResp.c_str());
+        kodi::QueueNotification(QUEUE_INFO, "Recording", "Recording Created");
+        kodi::addon::CInstancePVRClient::TriggerTimerUpdate();
+        return PVR_ERROR_NO_ERROR;
+      }else{
+        // record series
+        kodi::Log(ADDON_LOG_DEBUG, "[add timer] Record single tag;");
+        // {"title":"Das A-Team","channel":"RTLNITRO"}
+        string postData = "{\"title\": \"" + timer.GetTitle() +
+                          "\",\"channel\":\"" + channel.waipuID + "\"" + "}";
+        string recordResp = HttpPost("https://recording-scheduler.waipu.tv/api/serials", postData, {{"Content-Type", "application/vnd.waipu.recording-scheduler-serials-v1+json"}});
+        kodi::Log(ADDON_LOG_DEBUG, "[add timer] repeating response: %s;", recordResp.c_str());
+        kodi::QueueNotification(QUEUE_INFO, "Recording", "Rule Created");
+        kodi::addon::CInstancePVRClient::TriggerRecordingUpdate();
+        kodi::addon::CInstancePVRClient::TriggerTimerUpdate();
+        return PVR_ERROR_NO_ERROR;
+      }
     }
   }
   return PVR_ERROR_FAILED;
